// Author: Pete Kirkham
//
// quad store matching query bench.
//
#include <xmmintrin.h>
#include <iostream>

#include "utils/core_utils.h"
#include "utils/string_utils.h"

__m128i* match2_x4 (int value0, int value1, int value2, int value3, __m128i* data, __m128i* result, size_t size);
__m128i* match2_x4_no_sse (int value0, int value1, int value2, int value3, __m128i* data, __m128i* result, size_t size);
void dump (__m128i* data, size_t limit) ;
void ndump (__m128i* data, size_t limit) ;


int main (int narg, char** argvp)
{
    size_t num_quads(1 * MiB);

    if (narg > 1)
    {
        num_quads = atoi(argvp[1]) * MiB;
    }

    size_t num_passes(16);

    if (narg > 2)
    {
        num_passes = atoi(argvp[2]);
    }

    std::cout << "number of quads: " << num_quads << std::endl;
    std::cout << "memory required: " << (2 * num_quads * 16) / MiB << " MiB" << std::endl;

    // create num_quads quads
    __m128i* data = new __m128i[num_quads];
    __m128i* result = new __m128i[num_quads];
    __m128i* result_tail(result);

    // fill in data
    for (size_t i(0); i < num_quads; ++i)
    {
        _mm_storeu_si128(data+i, _mm_set_epi32((i&0xff)^0x34, ((i>>2)&0xff)^0xd0, ((i>>4)&0xff)^0x52, 0xffffffff));
        _mm_storeu_si128(result+i, _mm_set_epi32(0xffffffff, 0xffffffff, 0xffffffff, 0));
    }

    // dump first 16 data
    // dump(data, 16);

    const char* sse_mode(narg > 3 ? argvp[3] : "");

    // match data
    if (strcmp(sse_mode, "no_sse"))
    {
        for (size_t pass(0); pass < num_passes; ++pass)
        {
            result_tail = match2_x4(210, 211, 212, 213, data, result, num_quads);
            std::cout << "pass " << pass << std::endl;
        }
    }
    else
    {
        for (size_t pass(0); pass < num_passes; ++pass)
        {
            result_tail = match2_x4_no_sse(210, 211, 212, 213, data, result, num_quads);
            std::cout << "pass " << pass << std::endl;
        }
    }

    // output result
    size_t nresults(size_t(result_tail - result));

    if (nresults > 16)
        std::cout << nresults << " results, first 16 printed: " << std::endl;
    else
        std::cout << nresults << " results: " << std::endl;

    ndump(result, 16);

    delete [] result;
    delete [] data;

    std::cout << "done." << std::endl;
}

// matches the second value in quad
__m128i* match2_x4 (int value0, int value1, int value2, int value3, __m128i* data, __m128i* result, size_t size)
{
    __m128i mask(_mm_set_epi32(value0, value1, value2, value3));

    for (size_t i(0); i < size; ++i)
    {
        __m128i in(_mm_loadu_si128(data+i));
        __m128i cmp_x4(_mm_cmpeq_epi32(_mm_shuffle_epi32(in, SPREAD_EPI32(1)), mask));
        __m128i cmp(_mm_or_si128(
                                _mm_or_si128(_mm_shuffle_epi32(cmp_x4, SPREAD_EPI32(0)), _mm_shuffle_epi32(cmp_x4, SPREAD_EPI32(1))),
                                _mm_or_si128(_mm_shuffle_epi32(cmp_x4, SPREAD_EPI32(2)), _mm_shuffle_epi32(cmp_x4, SPREAD_EPI32(3)))));

        if (_mm_extract_epi16(cmp, 0))
        {
            _mm_storeu_si128(result, in);
            ++result;
        }
    }

    return result;
}

// matches the second value in quad without using sse2
__m128i* match2_x4_no_sse (int value0, int value1, int value2, int value3, __m128i* data, __m128i* result, size_t size)
{
    int*    idata((int*)data);

    for (size_t i(0), imax(size*4); i < imax; i += 4)
    {
        int value(idata[i + 2]);
        if ((value == value0) || (value == value1) || (value == value2) || (value == value3))
        {
            _mm_storeu_si128(result, _mm_loadu_si128(data+(i/4)));
            ++result;
        }
    }

    return result;
}


// dumps the data
void dump (__m128i* data, size_t limit)
{
    for (size_t i(0); i < limit; ++i)
        std::cout << to_i4(_mm_loadu_si128(data+i)) << std::endl;
}

// dumps the data until first null
void ndump (__m128i* data, size_t limit)
{
    for (size_t i(0); i < limit; ++i)
        if (!(*((int*)(&data[i]))))
            break;
        else
            std::cout << to_i4(_mm_loadu_si128(data+i)) << std::endl;
}
